import os
import PyMockTurtle
import pytest
import serial
import time

@pytest.fixture
def trtl_binary_hmq_purge(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-purge/fw-hmq-purge.bin")

@pytest.fixture
def trtl_binary_hmq_async_send(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-async-send/fw-hmq-async-send.bin")

@pytest.fixture
def trtl_binary_hmq_async_recv(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-async-recv/fw-hmq-async-recv.bin")

@pytest.fixture
def trtl_binary_hmq_sync(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-sync/fw-hmq-sync.bin")


class TestHmq(object):
    confirm = 'OK\r\n'

    @pytest.mark.parametrize("pattern,nmsg", [(0x11111111, 1),
                                              (0x88888888, 8),
                                              (0xAAAAAAAA, 16),
                                              ])
    def test_purge(self, trtl_cpu, trtl_binary_hmq_purge, pattern, nmsg):
        msg = PyMockTurtle.TrtlMessage()
        msg.header.len = 1
        msg.payload[0] = pattern

        filename = "/dev/ttytrtl-{:04x}-{:d}".format(trtl_cpu.trtl_dev.device_id,
                                                     trtl_cpu.idx_cpu)
        with serial.Serial(filename) as ser:
            ser.reset_input_buffer()

            trtl_cpu.load_application_file(trtl_binary_hmq_purge)
            trtl_cpu.enable()

            for hmq in trtl_cpu.hmq:
                for n in range(nmsg):
                    hmq.send_msg(msg)

            time.sleep(1)

            assert ser.in_waiting >= len(self.confirm)
            assert self.confirm == ser.read(ser.in_waiting).decode()

    def test_async_send(self, trtl_cpu, trtl_msg, trtl_binary_hmq_async_send):
        """It sends the test messages on all available HMQ.

        It validates:
        - messages are correctly sent
        - statistic counter increase coherently

        The test is successful when we read "OK" from the serial console
        """
        trtl_cpu.load_application_file(trtl_binary_hmq_async_send)
        filename = "/dev/ttytrtl-{:04x}-{:d}".format(trtl_cpu.trtl_dev.device_id,
                                                     trtl_cpu.idx_cpu)
        with serial.Serial(filename) as ser:
            ser.reset_input_buffer()
            trtl_cpu.enable()

            for hmq in trtl_cpu.hmq:
                for msg in trtl_msg:
                    sb = hmq.get_stats()
                    hmq.send_msg(msg)
                    sa = hmq.get_stats()
                    assert sb["message_sent"] + 1 == sa["message_sent"]

            time.sleep(0.1)

            assert ser.in_waiting >= len(self.confirm)
            assert self.confirm == ser.read(ser.in_waiting).decode()

    def test_async_recv(self, trtl_cpu, trtl_binary_hmq_async_recv):
        """It receives messages on all available HMQ.

        It validates:
        - messages are correctly received
        - statistic counter increase coherently

        The test is succesful if the received message content is the expected
        one
        """
        sb = {}
        for hmq in trtl_cpu.hmq:
            sb[hmq] = hmq.get_stats()

        trtl_cpu.load_application_file(trtl_binary_hmq_async_recv)
        trtl_cpu.enable()
        time.sleep(0.1)

        for hmq in trtl_cpu.hmq:
            tot = trtl_cpu.trtl_dev.rom.hmq[trtl_cpu.idx_cpu][hmq.idx_hmq].entries
            for n in range(tot):
                msg = hmq.recv_msg()
                payload = list(msg.payload)
                for i, val in enumerate(payload):
                    if i >= msg.header.len:
                        break
                    assert i == val

            sa = hmq.get_stats()
            assert sb[hmq]["message_received"] + tot == sa["message_received"]

    def test_sync(self, trtl_cpu, trtl_msg, trtl_binary_hmq_async_send):
        """It sends the test messages on all available HMQ.
        The test is successful when what we read back is the same we sent
        """
        trtl_cpu.load_application_file(trtl_binary_hmq_async_send)
        trtl_cpu.enable()

        for hmq in trtl_cpu.hmq:
            for msg in trtl_msg:
                sb = hmq.get_stats()
                msg_r = hmq.sync_msg(msg)
                sa = hmq.get_stats()
                assert sb["message_sent"] + 1 == sa["message_sent"]
                assert sb["message_received"] + 1 == sa["message_received"]

                assert msg_r.header.rt_app_id == msg.header.rt_app_id
                assert msg_r.header.msg_id == msg.header.msg_id
                assert msg_r.header.len == msg.header.len
                assert msg_r.header.sync_id == msg.header.sync_id
                assert msg.header.flags == PyMockTurtle.TrtlHmqHeader.TRTL_HMQ_HEADER_FLAG_SYNC
                assert msg_r.header.flags == PyMockTurtle.TrtlHmqHeader.TRTL_HMQ_HEADER_FLAG_ACK
                for v1, v2 in zip(msg.payload, msg_r.payload):
                    assert v1 == v2
